/*
** Module   : ORMHandler.java
** Abstract : Manages loading of ORM documents into a Hibernate Configuration
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20080305   @37508 Created initial version. Manages the loading
**                           of Hibernate ORM documents into a Hibernate
**                           Configuration.
** 002 GES 20080318   @37641 Switched from writer to stream to avoid
**                           unwanted character conversion when the
**                           default character set is not 8-bit ASCII.
** 003 CA  20080425   @38113 Added possibility to remove the foreign-key
**                           mappings from the hbm files, for the dirty
**                           databases.
** 004 ECF 20080509   @38497 Minor change to logging. Added database info
**                           to HBM dump.
** 005 GES 20090515   @42216 Import changes.
** 006 CA  20101026          Made the class and c'tor public. Built a new 
**                           mapClass method which takes a list of indexed, 
**                           computed columns.
** 007 SVL 20120331          Upgraded to Hibernate 4. Now this class stores mappings instead of
**                           immediately applying it to Hibernate configuration.
** 008 VMN 20131012          Added injectComputedColumns.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.cfg.*;
import org.hibernate.internal.util.xml.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import com.goldencode.util.*;
import com.goldencode.p2j.persist.dialect.*;

/**
 * Class which manages the loading of a Hibernate <code>Configuration</code>.
 * Each ORM is loaded via the {@link #mapClass} method.  Depending upon the
 * database dialect, this method may inject specialized nodes into the ORM XML
 * documents before handing them off to Hibernate.  This is how we manage the
 * use of computed columns for databases which do not allow expressions within
 * indexes.
 */
public class ORMHandler
{
   /** Foreign-key column name suffix */
   static final String FK_SUFFIX = "__fk";
   
   /** Logger */
   private static final Log LOG = LogFactory.getLog(ORMHandler.class);
   
   /** XML property element */
   private static final String ELEM_PROPERTY = "property";
   
   /** XML many-to-one element */
   private static final String ELEM_MANY_TO_ONE = "many-to-one";

   /** XML one-to-one element */
   private static final String ELEM_ONE_TO_ONE = "one-to-one";

   /** XML name attribute */
   private static final String ATTR_NAME = "name";
   
   /** XML column attribute */
   private static final String ATTR_COLUMN = "column";
   
   /** XML access attribute */
   private static final String ATTR_ACCESS = "access";
   
   /** XML insert attribute */
   private static final String ATTR_INSERT = "insert";
   
   /** XML update attribute */
   private static final String ATTR_UPDATE = "update";
   
   /** NopPropertyAccessor fully qualified class name */
   private static final String NOP_ACCESSOR =
      "com.goldencode.p2j.persist.NopPropertyAccessor";
   
   /** P2J database */
   private final Database database;
   
   /** Prefix to prepend on computed column property */
   private final String computedPrefix;
   
   /** Entity resolver used when parsing XML */
   private final EntityResolver entityResolver;
   
   /** This flag determines if the foreign keys should be removed or not */
   private boolean removeForeignKeys = false;

   /** Instance of XML helper. */
   private final XMLHelper xmlHelper;

   /** Stored Hibernate mappings. */
   private final Set<XmlDocument> hibernateMappings = new HashSet<XmlDocument>();

   /** Instance of entity resolver. */
   private EntityResolver streamEntityResolver = null;

   /** This flag determines if computed columns should be injected on table creating or not */
   private boolean injectComputedColumns;

   /**
    * Constructor.
    * 
    * @param   dialect
    *          P2J database dialect.
    * @param   database
    *          Database whose ORM mappings require loading.
    */
   public ORMHandler(P2JDialect dialect,
                     Database database)
   {
      this.removeForeignKeys = database.isDirty();
      this.database = database;
      this.xmlHelper = new XMLHelper();
      this.injectComputedColumns = dialect.injectComputedColumns();

      if (dialect.needsComputedColumns() && injectComputedColumns)
      {
         computedPrefix = dialect.getComputedColumnPrefix();
         entityResolver = new DTDEntityResolver();
      }
      else
      {
         computedPrefix = null;

         if (removeForeignKeys)
         {
            entityResolver = new DTDEntityResolver();
         }
         else
         {
            entityResolver = null;
            streamEntityResolver = new EJB3DTDEntityResolver();
         }
      }
   }

   /**
    * Apply specified Hibernate mappings to the given Hibernate configuration.
    *
    * @param mappings
    *        Hibernate mappings to apply (a set of loaded XML documents).
    *
    * @param configuration
    *        Hibernate configuration to which the mappings are applied.
    */
   public static void applyHibernateMappings(Set<XmlDocument> mappings,
                                             Configuration configuration)
   {
      for (XmlDocument document : mappings)
      {
         configuration.add(document);
      }
   }

   /**
    * Get stored Hibernate mappings.
    *
    * @return  Stored Hibernate mappings.
    */
   public Set<XmlDocument> getHibernateMappings()
   {
      return hibernateMappings;
   }

   /**
    * Configure the specified DMO class, updating the object to relational
    * mapping to inject computed column property nodes if the database dialect
    * requires it.
    * 
    * @param   schema
    *          Database schema.
    * @param   dmoClass
    *          DMO implementation class.
    * 
    * @throws  PersistenceException
    *          if there is an error injecting computed column property nodes
    *          into the ORM XML.
    */
   void mapClass(String schema, Class<?> dmoClass)
   throws PersistenceException
   {
      Set<String> indexedCols = null;
      
      if (computedPrefix != null)
      {
         Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
         
         if (!DatabaseManager.isPermamentDB(database))
         {
            indexedCols = TempTableHelper.getComputedCols(database, dmoIface)
                                         .keySet();
         }
         else
         {
            // for permanent tables, we must read the indexes from the 
            // underlying database using JDBC.
            String table = DatabaseManager.getBackingTableName(dmoClass);
            indexedCols = DMOIndex.getIndexedCharacterColumns(schema,
                                                              database,
                                                              table)
                                  .keySet();
         }
      }

      String ormLoc = getORMLoc(dmoClass);
      ClassLoader loader = dmoClass.getClassLoader();
      InputStream in = loader.getResourceAsStream(ormLoc);
      
      mapClass(schema, in, ormLoc, indexedCols);
   }

   /**
    * Configure the DMO class by reading its definition from the specified
    * stream, updating the object to relational mapping to inject computed 
    * column property nodes if the database dialect requires it.
    * 
    * @param   schema
    *          Database schema.
    * @param   in
    *          The ORM input stream.
    * @param   ormLoc
    *          The ORM file location.
    * @param   indexedCols
    *          A set containing the names of the indexed character columns.
    * 
    * @throws  PersistenceException
    *          if there is an error injecting computed column property nodes
    *          into the ORM XML.
    */
   void mapClass(String schema, 
                 InputStream in,
                 String ormLoc,
                 Set<String> indexedCols)
   throws PersistenceException
   {
      if (computedPrefix == null || !injectComputedColumns)
      {
         if (removeForeignKeys)
         {
            Document doc = removeForeignKeys(in, ormLoc);
            storeMappingDoc(doc);
         }
         else
         {
            storeMappingStream(in);
         }
      }
      else
      {
         // after injected the computed property nodes, the foreign keys will 
         // be removed if needed
         injectComputedColumnsAndProcessFK(indexedCols, in, ormLoc);
      }
   }

   /**
    * Configure the DMO class by reading its definition from the specified ORM
    * file, updating the object to relational mapping to inject computed 
    * column property nodes if the database dialect requires it.
    * 
    * @param   schema
    *          Database schema.
    * @param   ormLoc
    *          The ORM file location.
    * @param   indexedCols
    *          A set containing the names of the indexed character columns.
    * 
    * @throws  PersistenceException
    *          if there is an error injecting computed column property nodes
    *          into the ORM XML.
    * @throws  FileNotFoundException
    *          if the ORM file was not found. 
    */
   public void mapClass(String schema, 
                        String ormLoc,
                        Set<String> indexedCols)
   throws PersistenceException, 
          FileNotFoundException
   {
      InputStream in = new FileInputStream(ormLoc);
      mapClass(schema, in, ormLoc, indexedCols);
   }
   
   /**
    * For each DMO character property which corresponds with a column
    * participating in a database index, inject a special property node into
    * the ORM XML document for the given class.  The injected property node
    * represents a read-only (no insert, no update) property, which does not
    * actually exist in the DMO.  Since there is no actual, corresponding
    * property in the DMO, a {@link NopPropertyAccessor} is used to intercept
    * calls to get and set the property's value in the DMO.  The element is
    * injected directly after the real property node.  The injected element
    * uses the same name as the real property, only with a leading prefix
    * prescribed by the corresponding database dialect.
    * <p>
    * Also, if the {@link #removeForeignKeys} flag is set to <code>true</code>,
    * all foreign key mappings are removed from the associated hbm.xml file.
    * 
    * @param   indexedColumns
    *          Set of those character column names which correspond with
    *          columns participating in database indexes. 
    * @param   in
    *          The ORM input stream.
    * @param   ormLoc
    *          The ORM file location.
    * 
    * @throws  PersistenceException
    *          if there is an error injecting computed column property nodes
    *          into the ORM XML.
    */
   private void injectComputedColumnsAndProcessFK(Set<String> indexedColumns,
                                                  InputStream in,
                                                  String ormLoc)
   throws PersistenceException
   {
      try
      {
         Document doc = loadHBMFile(in);
         Element root = doc.getDocumentElement();
         
         walkChildren(indexedColumns, root);
         
         if (removeForeignKeys)
         {
            removeForeignKeys(root);
         }
         
         if (LOG.isDebugEnabled())
         {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            XmlHelper.write(doc, baos, true);
            String xml = new String(baos.toByteArray(), "UTF-8");
            
            StringBuilder buf = new StringBuilder();
            buf.append("Preprocessed ORM [");
            buf.append(database);
            buf.append("] ");
            buf.append(ormLoc);
            buf.append("):\n");
            buf.append(xml);
            
            LOG.debug(buf.toString());
         }
         
         storeMappingDoc(doc);
      }
      catch (Exception exc)
      {
         throw new PersistenceException(exc);
      }
   }

   /**
    * This method gets the ORM file location for the given DMO class.
    * 
    * @param   dmoClass
    *          The DMO class for which to load and parse the hbm.xml file.
    * 
    * @return  the ORM file location.
    */
   private String getORMLoc(Class<?> dmoClass)
   {
      // Build qualified mapping document filename.
      String dmoClsName = dmoClass.getName();
      String ormLoc = dmoClsName.replace('.', '/');
      StringBuilder buf = new StringBuilder(ormLoc);
      buf.append(".hbm.xml");
      ormLoc = buf.toString();
      
      return ormLoc;
   }

   /**
    * This method loads and parses the specified hbm.xml.
    * 
    * @param   in
    *          The ORM stream.
    *          
    * @return  XML document created by parsing the associated hbm.xml file.
    * 
    * @throws  Exception
    *          If could not load or parse the hbm.xml file.
    */
   private Document loadHBMFile(InputStream in)
   throws Exception
   {     
      Document doc = XmlHelper.parse(in, true, entityResolver, null);    

      return doc;
   }
   
   /**
    * Remove the foreign keys from the associated hbm.xml ORM file associated
    * with the given DMO, and return the document.
    * 
    * @param   in
    *          The ORM input stream.
    * @param   ormLoc
    *          The ORM file location.
    * 
    * @return  the associated XML document with all the foreign keys removed.
    * 
    * @throws  PersistenceException
    *          if there is an error removing the foreign key nodes from the
    *          ORM XML.
    */
   private Document removeForeignKeys(InputStream in, String ormLoc)
   throws PersistenceException
   {
      try
      {
         Document doc = loadHBMFile(in);
         Element root = doc.getDocumentElement();
         
         removeForeignKeys(root);
         
         if (LOG.isDebugEnabled())
         {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            XmlHelper.write(doc, baos, true);
            String xml = new String(baos.toByteArray(), "UTF-8");
            
            StringBuilder buf = new StringBuilder();
            buf.append("Preprocessed ORM [");
            buf.append(database);
            buf.append("] ");
            buf.append(ormLoc);
            buf.append("):\n");
            buf.append(xml);
            
            LOG.debug(buf.toString());
         }

         return doc;
      }
      catch (Exception exc)
      {
         throw new PersistenceException(exc);
      }
   }
   
   /**
    * This method walks the DOM tree and removes the following nodes, when
    * a node with a <code>column</code> attribute containing a value which 
    * ends with the <code>__fk</code> suffix is encountered:
    * <ul>
    *    <li><code>many-to-one</code> nodes
    *    <li><code>one-to-one</code> nodes
    *    <li>if there is a <code>column</code> attribute with its value ending
    *        with the <code>__fk</code> suffix, remove the parent node if 
    *        this node is not a <code>many-to-one</code> or <code>one-to-one
    *        </code> node; this ensures that all collections - set, map, etc -
    *        are removed.
    * </ul>
    *  
    * @param   parent
    *          The parent node from which to start removing the foreign-key 
    *          nodes.
    * 
    * @return  <code>true</code> if the <code>parent</code> node should be 
    *          removed.
    */
   private boolean removeForeignKeys(Node parent)
   {
      Node child = parent.getFirstChild();
      Node nextChild = null;
      
      boolean removeParent = false;
      
      while (child != null)
      {
         boolean remove = removeForeignKeys(child);
         
         // if this node should be removed (it represents a collection - map,
         // set, etc) then remove it and continue processing
         if (remove)
         {
            nextChild = child.getNextSibling();
            parent.removeChild(child);
            child = nextChild;
            continue;
         }
         
         remove = false;
         
         // skip non-nodes elements
         short nodeType = child.getNodeType();
         if (nodeType != Node.ELEMENT_NODE)
         {
            // do nothing
         }
         else
         {
            Element elem = (Element) child;
            
            String column = elem.getAttribute(ATTR_COLUMN);
            String tag = elem.getTagName();
            
            // if this is a one-to-one relation, remove the node
            if (tag.equals(ELEM_ONE_TO_ONE))
            {
               remove = true;
            }
            else
            {
               if (column != null && column.endsWith(FK_SUFFIX))
               {
                  // if this is a many-to-one relation, remove the node
                  if (tag.equals(ELEM_MANY_TO_ONE))
                  {
                     remove = true;
                  }
                  else
                  {
                     // else, if this is not a many-to-one relation, the only 
                     // possible conclusion is that this is a collection (map,
                     // set, etc) - so remove the parent node
                     removeParent = true;
                  }
               }
            }
         }
         
         if (remove)
         {
            nextChild = child.getNextSibling();
            parent.removeChild(child);
            child = nextChild;
         }
         else
         {
            child = child.getNextSibling();
         }
      }
      
      return removeParent;
   }
   
   /**
    * Walk the children of an XML element node, testing each to determine
    * whether it represents a property which corresponds with a text column
    * which participates in a database index.  For those that test positive,
    * inject a computed a computed column property element after them.
    * <p>
    * This method calls itself recursively.
    * 
    * @param   indexedColumns
    *          Set of those character column names which correspond with
    *          columns participating in database indexes. 
    * @param   node
    *          XML parent node whose children are to be visited.
    */
   private void walkChildren(Set<String> indexedColumns, Node node)
   {
      Node next = node.getFirstChild();
      Node injected = null;
      while (next != null)
      {
         // Analyze child node and possibly inject a computed column entry
         // after it.
         injected = visitChild(indexedColumns, node, next);
         
         // Recursively walk children, depth-first.
         walkChildren(indexedColumns, next);
         
         // Get next node.
         next = next.getNextSibling();
         
         // Skip over an injected node.
         if (injected != null && next == injected)
         {
            next = next.getNextSibling();
         }
      }
   }
   
   /**
    * Visit an XML node and possibly inject a computed column property element
    * directly after it.  The node must meet the following conditions for the
    * element to be injected behind it:
    * <ul>
    *   <li>it must be an element node;
    *   <li>it must have a 'property' tag name;
    *   <li>it must reference a database text column which participates in an
    *       index.
    * </ul>
    * 
    * @param   indexedColumns
    *          Set of those character column names which correspond with
    *          columns participating in database indexes. 
    * @param   parent
    *          Visited node's parent node.
    * @param   child
    *          XML node being visited.
    * 
    * @return  Injected XML node, or <code>null</code> if no node was
    *          injected.
    */
   private Node visitChild(Set<String> indexedColumns,
                           Node parent,
                           Node child)
   {
      short nodeType = child.getNodeType();
      if (nodeType != Node.ELEMENT_NODE)
      {
         return null;
      }
      
      Element elem = (Element) child;
      String tagName = elem.getTagName();
      if (!ELEM_PROPERTY.equals(tagName))
      {
         return null;
      }
      
      String colName = elem.getAttribute(ATTR_COLUMN);
      if (!indexedColumns.contains(colName))
      {
         return null;
      }
      
      String propName = elem.getAttribute(ATTR_NAME);
      StringBuilder buf = new StringBuilder(computedPrefix);
      buf.append(propName);
      propName = buf.toString();
      
      buf = new StringBuilder(computedPrefix);
      buf.append(colName);
      colName = buf.toString();
      
      Element copy = (Element) elem.cloneNode(false);
      
      copy.setAttribute(ATTR_NAME, propName);
      copy.setAttribute(ATTR_COLUMN, colName);
      copy.setAttribute(ATTR_INSERT, "false");
      copy.setAttribute(ATTR_UPDATE, "false");
      copy.setAttribute(ATTR_ACCESS, NOP_ACCESSOR);
      
      // Insert before current node.
      parent.insertBefore(copy, child);
      
      // Invert the order of these two nodes, so computed column comes last.
      parent.insertBefore(child, copy);
      
      return copy;
   }

   /**
    * Read the Hibernate mappings from the given input stream and store them.
    *
    * @param inputStream
    *        Input stream containing Hibernate mappings.
    */
   private void storeMappingStream(InputStream inputStream)
   {
      InputSource inputSource = new InputSource(inputStream);
      Origin origin = new OriginImpl("input stream", null);
		try
      {
			XmlDocument xml =
               MappingReader.INSTANCE.readMappingDocument(streamEntityResolver,
                                                          inputSource,
                                                          origin);
         hibernateMappings.add(xml);
		}
		finally
      {
			try
         {
				inputStream.close();
			}
			catch (IOException ignore)
         {
				LOG.trace( "Was unable to close input stream");
			}
		}
   }

   /**
    * Read the Hibernate mappings from the given document and store them.
    *
    * @param doc
    *        Document containing Hibernate mappings.
    */
   private void storeMappingDoc(Document doc)
   {
      org.dom4j.Document document = xmlHelper.createDOMReader().read(doc);
      hibernateMappings.add(new XmlDocumentImpl(document, "unknown", null));
   }
}
