/*
** Module   : P2JSQLServer2012Dialect.java
** Abstract : Implementation of P2JDialect for a SQL Server 2012 backend
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 VMN 20131012 Created initial version with conversion time sequences support.
*/

package com.goldencode.p2j.persist.dialect;

import org.apache.commons.logging.*;
import com.goldencode.p2j.persist.sequence.*;

/**
 * Extends P2JSQLServer2008Dialect to add method implementations supported with SQL Server 2012.
 * <p>
 * Note:  this implementation expects a directory to be available at the time
 * the class is loaded.
 */
public class P2JSQLServer2012Dialect
extends P2JSQLServer2008Dialect
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(P2JSQLServer2012Dialect.class);

   /** Debug level logging enabled? */
   private static final boolean debug = LOG.isDebugEnabled();

   /**
    * Generate a DDL string which will create a global sequence with the given
    * starting value.
    *
    * @param   name
    *          Sequence name.
    * @param   start
    *          Sequence starting value.
    *
    * @return  Dialect-specific string to create the sequence.
    */
   public String getCreateSequenceString(String name, long start)
   {
      StringBuilder buf = new StringBuilder("create sequence ");
      buf.append(name);
      buf.append(" start with ");
      buf.append(start);

      return buf.toString();
   }

   /**
    * Generate a DDL string which will create a sequence with given properties.
    *
    * @param   name
    *          Sequence name. Mandatory not null.
    * @param   init
    *          Sequence starting value. Mandatory(not null).
    * @param   increment
    *          Sequence incrementing value. Mandatory(not null).
    * @param   minVal
    *          Sequence minimum value. Optional (may be null).
    * @param   maxVal
    *          Sequence maximum value. Optional (may be null).
    * @param   cycle
    *          Sequence cycling property. Optional, if missing assumed not cycling.
    *
    * @return  Dialect-specific string to create the sequence.
    */
   public String getCreateSequenceString(String name,
                                         Long init,
                                         Long increment,
                                         Long minVal,
                                         Long maxVal,
                                         Boolean cycle)
   {
      StringBuilder buf = new StringBuilder("create sequence ");
      buf.append(name);
      buf.append(" increment ").append(increment);

      // START WITH <constant>
      // The START value must be a value less than or equal to the maximum and greater than or
      // equal to the minimum value of the sequence object. The default start value for a new
      // sequence object is the minimum value for an ascending sequence object and the maximum
      // value for a descending sequence object.
      //
      // MAXVALUE.
      // The default minimum value for a new sequence object is the minimum value of the data
      // type of the sequence object. This is zero for the tinyint data type and a negative
      // number for all other data types.

      // MAXVALUE.
      // The default maximum value for a new sequence object is the maximum value of the data
      // type of the sequence object.

      if (minVal != null)
      {
         buf.append(" minvalue ").append(minVal);
         if (increment > 0 && init < minVal)
         {
            init = minVal;
         }
      }

      if (maxVal != null)
      {
         buf.append(" maxvalue ").append(maxVal);
         if (increment < 0 && init > maxVal)
         {
            init = maxVal;
         }
      }

      buf.append(" start with ").append(init);
      if (cycle != null && cycle)
      {
         buf.append(" cycle");
      }
      // TODO cache parameter is also supported but:
      // If the cache option is enabled without specifying a cache size, the Database Engine
      // will select a size. However, users should not rely upon the selection being consistent.
      // Microsoft might change the method of calculating the cache size without notice.

      return buf.toString();
   }

   /**
    * Generate a DDL string which will drop a sequence with given properties.
    *
    * @param   name
    *          The name of the sequence to be dropped.
    *
    * @return  Dialect-specific string to drop the sequence.
    */
   public String getDropSequenceString(String name)
   {
      // http://technet.microsoft.com/en-us/library/ff877934.aspx and
      // http://technet.microsoft.com/en-us/library/ms190324.aspx
      // type SO = Sequence object
      StringBuilder buf = new StringBuilder(
         "if exists (select * from sys.sequences where name = '");
      buf.append(name);
      buf.append("' and type='SO') drop sequence ");
      buf.append(name);

      return buf.toString();
   }

   /**
    * Generate the appropriate select statement to retrieve the next value of a sequence.
    *
    * @param   sequenceName
    *          The name of the sequence.
    * @param   newVal
    *          The new value to be set.
    *
    * @return  Dialect-specific string to query the sequence.
    */
   public String getSequenceSetValString(String sequenceName, long newVal)
   {
      StringBuilder sb = new StringBuilder("alter sequence ");
      sb.append(sequenceName);
      sb.append(" restart with ");
      sb.append(newVal);

      return sb.toString();
   }

   /**
    * Generate the appropriate select statement to reset the value of a sequence.
    *
    * @param   sequenceName
    *          The name of the sequence to be queried.
    *
    * @return  Dialect-specific string to query the sequence.
    */
   public String getSequenceCurrValString(String sequenceName)
   {
      StringBuilder buf = new StringBuilder("next value for ");
      buf.append(sequenceName);
      buf.append("'");

      return buf.toString();
   }

   /**
    * Returns <code>true</code> if this dialect fully support the sequences as
    * 4GL language (most important: bounded values, cycle). If not fully
    * supported, additional code is needed in P2J sequence class
    * implementation to fix this.
    *
    * @return  <code>true</code> because Server 2012 supports the concept of a SEQUENCE.
    */
   public boolean hasFullSequenceSupport()
   {
      return true;
   }

   /**
    * Creates and returns a <code>SequenceHandler</code> for a SQL Server 2012 database.
    *
    * @param   ldbName
    *          The logical name of the database to create the <code>SequenceHandler</code> for.
    *
    * @return  A <code>SQLServer2012SequenceHandler</code> instance.
    */
   @Override
   public SequenceHandler createSequenceHandler(String ldbName)
   {
      return new SQLServer2012SequenceHandler(ldbName);
   }

}
